# -*- coding: utf-8 -*-
"""
Created on Fri Apr 28 15:42:38 2017

@author: gjtim
"""
# Imports
from matplotlib.path import Path
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import math
import matplotlib.path as mpath
import matplotlib.image as mpimg
import matplotlib.animation as manimation
import matplotlib.font_manager as fm

FFMpegWriter = manimation.writers['ffmpeg']
metadata = dict(title='Electric or Magnetic?', artist='',comment='')
writer = FFMpegWriter(fps=24, metadata=metadata)


pi = 3.14159

# Set canvas size in inches
#pltwid = 12       # For 12 inches
#pltht = 8
pltwid = 19.2       # For 1920 x 1080 HD video
pltht = 10.8
margin = 1.6
lm = margin
rm = pltwid - margin
bm = margin
tm = pltht - margin 

# Variables for wire etc
wdia = 1.8     # Wire diameter
ww = 2        # Line width
xl = margin
xr = pltwid - margin
yb = bm + 2.5
yt = yb + wdia
yc = (yt+yb)/2
posanph = 0              # +ve animation phase
neganph = 0
battery_current = 0
chgsize = 0.15
lwchg = 2
ycharge = yc + 3.5     # Y pos of moving charge
xcharge = lm
dxchg = 0.08            # dx for charge step
chargespacing = 0.33     # Normal separation of charges
spposx = chargespacing
spnegx = spposx
rel_shrink = 0.3          # Relativity shrinkage factor: 0.4 for 40% shrink
chargespeed = 0.3*chargespacing    # Movement per time step
zf = 0                  # Frame counter
xarrow = lm

frames1 = 200
frames2 = 200
frames3 = 200
frames4 = 200
frames5 = 120
frames6 = 200
frames7 = 200

#frames1 = 0
#frames2 = 0
#frames3 = 0
#frames4 = 0
#frames5 = 0
#frames6 = 20
#frames7 = 20

def setplot():
# ************** Initialize the canvas *****************************
    plt.xlim(0,pltwid)      # Set x limits
    plt.xticks(np.linspace(0,10,0,endpoint=True))   # Set x ticks
    # plt.xticks(np.linspace(start, finish,Number of ticks,endpoint+True))
    plt.ylim(0,pltht)       # Set y limits
    plt.yticks(np.linspace(-1,1,0,endpoint=True))   # Set y ticks
    plt.axis('off')         # axis off and tight make it all bigger
    plt.tight_layout()

fig = plt.figure(figsize=(pltwid,pltht))
setplot()     

''' ************ Functions *************** '''
def Write_Caption(flag,t):
    y = pltht - 0.3
    linespacing = 0.45
    if flag:
        y -= flag*linespacing
    plt.text(pltwid/2, y, t,fontsize=26, horizontalalignment='center')
    
def Draw_Wire():
    global xarrow
    plt.plot([xl, xr], [yb, yb], 'black', linestyle = '-', lw=ww)
    plt.plot([xl, xr], [yt, yt], 'black', linestyle = '-', lw=ww)    
    if wire_moving:
        # Draw moving arrow heads on wire sides
        arrow_spacing = 1
        headlen = 0.3
        headangle = 0.25
        xarrow += posanph
        #xarrow = xarrow + posanph
        x = xarrow
        for z in range(30):
            if x >= xl + headlen/2:
                # Draw arrow heads along wire sides
                ArrowHead(x,yb,pi,headlen,headangle)
                ArrowHead(x,yt,pi,headlen,headangle)
            x += arrow_spacing
            if x >= xr-headlen/2:
                break
            
def Blanck_Wire_Ends():
    # Blank ends to left and right of wire
    x = 0
    y = yc-wdia/2
    blanker = mpatches.Rectangle((x, y), xl, wdia, color='white', zorder=10)
    plt.gca().add_patch(blanker)
    x = pltwid-xl
    blanker = mpatches.Rectangle((x, y), xl, wdia, color='white', zorder=10)
    plt.gca().add_patch(blanker)        
    
def Draw_Charges(zf):
    # Draw lines of n +ve charges

    yp = yc - 1.8*chargespacing
    ym = yp - 0.5*chargespacing
    # Reference position must stay central, aligned with test charge
#    xpref = pltwid/2 + zf*posanph       # Add the animation offset
#    # Draw -ve charges
#    if battery_current:
#        xmref = pltwid/2 + (zf*neganph % chargespacing)       # Add the animation offset
#    else:
#        xmref = pltwid/2 + zf*neganph
    xpref = pltwid/2 + xarrow - lm       # Add the animation offset
    # Draw -ve charges
    if battery_current:
        xmref = pltwid/2 + (zf*neganph % chargespacing)       # Add the animation offset
    else:
        xmref = pltwid/2 + xarrow - lm

    
    for zy in range (5):
        x = xpref
        y = yp+zy*chargespacing
        # Draw +ve charges to right of centre
        while True:            
            #plt.text(x, yp+zy*chargespacing,'+',fontsize=16, color='red', verticalalignment='center', horizontalalignment='center')
            plt.plot([x-chgsize/2, x+chgsize/2], [y, y], 'red', lw=lwchg)
            plt.plot([x, x], [y-chgsize/2, y+chgsize/2], 'red', lw=lwchg)
            x += spposx
            if x >= xr:
                break

        x = xpref
        # Draw +ve charges to left of centre
        while True:            
            #plt.text(x, yp+zy*chargespacing,'+',fontsize=16, color='red', verticalalignment='center', horizontalalignment='center')
            plt.plot([x-chgsize/2, x+chgsize/2], [y, y], 'red', lw=lwchg)
            plt.plot([x, x], [y-chgsize/2, y+chgsize/2], 'red', lw=lwchg)
            x -= spposx
            if x <= xl:
                break
            
        x = xmref
        y = ym+zy*chargespacing        
        # Draw -ve charges to right of centre
        while True:            
            plt.plot([x-chgsize/2, x+chgsize/2], [y, y], 'blue', lw=lwchg)
            x += spnegx
            if x >= xr:
                break

        x = xmref
        # Draw -ve charges to right of centre
        while True:            
            plt.plot([x-chgsize/2, x+chgsize/2], [y, y], 'blue', lw=lwchg)
            x -= spnegx
            if x <= xl:
                break

            
            
def Draw_Test_Charge():
    # Draw Circle (white for blanking)
    circle = mpatches.Circle((xchg, ychg), 0.11, edgecolor='black', linewidth=1, facecolor='white')
    plt.gca().add_patch(circle)            
    plt.plot([xchg-chgsize/2, xchg+chgsize/2], [ychg+0.0035, ychg+0.0035], 'blue', lw=lwchg)     
    #plt.text(xchg-0.005, ychg+0.1,'_',fontsize=16, color='blue', verticalalignment='center', horizontalalignment='center')
            
def Draw_Field_Ellipses():
    # Draw ellipses to show the magnetic field
    n = 6
    x = xl + 1.4
    dx = (xr-xl)/n
    ang = 1.33 #1.26        
    headlen = 0.36
    headangle = 0.31
    headxoff = 0.425 #0.295
    headyoff = 1.65
    dtheta = 23.5
    for zr in range(n):
        arc = mpatches.Arc((x,yc), 0.7*wdia, 2.5*wdia, theta1=dtheta, theta2=360.0-dtheta, edgecolor='black', linewidth=ww)
        plt.gca().add_patch(arc)
        # Put Arrow Heads on magneti field lines
        ArrowHead(x-headxoff,yc+headyoff,ang,headlen,headangle)
        ArrowHead(x+headxoff,yc-headyoff,ang+pi,headlen,headangle)            
        x += dx

def Draw_Electric_Field():
    # Draw arrows for the electric field
    n = 6
    x = xl + 0.95        
    dx = (xr-xl)/n
    ylen = 1.4
    headlen = 0.36
    headangle = 0.31    
    for zr in range(n):
        plt.plot([x, x], [yc + wdia/2, yc + wdia/2 + ylen], 'black', lw=ww)
        ArrowHead(x,yc + 2.4, pi/2, headlen, headangle)
        plt.plot([x, x], [yc - wdia/2, yc - wdia/2 - ylen], 'black', lw=ww)
        ArrowHead(x,yc - 2.4, 3*pi/2, headlen, headangle)        
        x += dx
        
def ArrowHead(xc,yc,ang,headlen,headangle):
    # Draw an arrow head with centre at x,y)   ang in radians
    x = xc + 0.5*headlen*math.cos(ang)
    y = yc + 0.5*headlen*math.sin(ang)
    x1 = xc - 0.5*headlen*math.cos(ang-headangle)
    y1 = yc - 0.5*headlen*math.sin(ang-headangle)
    x2 = xc - 0.5*headlen*math.cos(ang+headangle)
    y2 = yc - 0.5*headlen*math.sin(ang+headangle)    
    verts = [
    (x, y),             # Start Position
    (x1, y1),             # End of 1st straight line
    (x2, y2),             # End of 2nd straight line
    (x, y),    
    ]
    codes = [Path.MOVETO,
         Path.LINETO,
         Path.LINETO,         
         Path.LINETO,
     ]
    path = mpath.Path(verts, codes)
    patch = mpatches.PathPatch(path, edgecolor='black', facecolor='black')
    plt.gca().add_patch(patch) 

def Draw_ZigZag(x,y,ht,linewid):
    # Draw zigzag ends of wire
    yz = ht/10
    xz = 0.06    
    verts = [
        (x, y),                  # Start Position
        (x-xz, y-0.5*yz),             # End of 1st straight line
        (x+xz, y-1.5*yz),
        (x-xz, y-2.5*yz),
        (x+xz, y-3.5*yz),
        (x-xz, y-4.5*yz),
        (x+xz, y-5.5*yz),
        (x-xz, y-6.5*yz),
        (x+xz, y-7.5*yz),
        (x-xz, y-8.5*yz),
        (x+xz, y-9.5*yz),
        (x, y-ht),
    ]
    codes = [Path.MOVETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,
         Path.LINETO,          
     ]
    path = mpath.Path(verts, codes)
    patch = mpatches.PathPatch(path, edgecolor='black', facecolor='white',lw=linewid ,zorder=10)
    plt.gca().add_patch(patch) 
    
def Draw_Circuit():
    # Draw battery and switch circuit
    xs = pltwid/2 - 0.8    
    ys = yc - 3.8      # Y coord of switch and battery
    cr = 1
    lwc = 3
    
    # Draw Arcs for each end of circuit
    archt = yc-ys
    carc = mpatches.Arc((rm, yc - archt/2), 0.7*archt, archt, theta1=270, theta2=90, edgecolor='black', linewidth=lwc, fill=False, zorder=10)
    plt.gca().add_patch(carc)
    carc = mpatches.Arc((lm, yc - archt/2), 0.7*archt, archt, theta1=90, theta2=270, edgecolor='black', linewidth=lwc, fill=False, zorder=10)
    plt.gca().add_patch(carc)
      
    # If wire shrunk, connect its ends
    if xl > lm:
        plt.plot([lm, xl], [yc, yc], 'black', lw=lwc,zorder=10)    
        plt.plot([xr, rm], [yc, yc], 'black', lw=lwc,zorder=10)
        
    # Wire to left of switch
    plt.plot([lm, xs], [ys, ys], 'black', lw=lwc)  
    circle = mpatches.Circle((xs, ys),cr/10, edgecolor='black', linewidth=lwc, facecolor='white',zorder=10)
    plt.gca().add_patch(circle)     
     
    # Wire from switch to battery
    x = xs+2*cr
    plt.plot([xs+1.1*cr, x], [ys, ys], 'black', lw=lwc)  # XXX - YYY !!    
    # Battery +ve
    plt.plot([x, x], [ys-cr/8, ys+cr/8], 'black', lw=lwc)
    x = x+cr/8
    # Battery -ve
    plt.plot([x, x], [ys-0.4*cr, ys+0.4*cr], 'black', lw=lwc)    
    # Wire to right of battery
    plt.plot([x, rm], [ys, ys], 'black', lw=lwc)
    
    # Draw RHS of switch
    circle = mpatches.Circle((xs+cr, ys), cr/10, edgecolor='black', linewidth=lwc, facecolor='white',)
    plt.gca().add_patch(circle)     
    if battery_current:
        plt.plot([xs+0.9*cr, xs], [ys, ys+0.1*cr], 'black', lw=lwc)
    else:
        plt.plot([xs+0.9*cr, xs], [ys, ys+0.45*cr], 'black', lw=lwc)        
    

        
''' Main Program '''
with writer.saving(fig, "Electric or Magnetic Animation.mp4", 100):    # The 100 is dpi

    ''' 1) Start: Wire still, charges still, no field  '''   
    wire_moving = False
    ychg = ycharge
    xchg = xcharge
    for zf in range(frames1):
    
        plt.clf()
        setplot()        
        if zf==0:
            print ()
            print ('1) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'A wire (central part enlarged) without an electric current contains')
        Write_Caption(1,'equal numbers of stationary +ve and -ve charges so it is neutral overall.')        
        Write_Caption(2,'The wire creates no electric or magnetic field.')        
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        Draw_Circuit()
        xchg += dxchg
        
        writer.grab_frame()

        
    ''' 2) Charge moving, Wire still, no current in wire, no field  '''   
    ychg = ycharge
    xchg = xcharge
    for zf in range(frames2):
        plt.clf()
        setplot()        
        if zf==0:
            print ()
            print ('2) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'A -ve test charge moving parallel to the wire continues in a straight line')
        Write_Caption(1,'because there are no electric or magnetic fields to affect it.')        
        #Write_Caption(2,'')        
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        #Draw_Field_Ellipses()
        Draw_Circuit()
        xchg += dxchg
        Draw_Test_Charge()
        
        writer.grab_frame()

        
    ''' 3) No Moving Charge, Wire still, Current in wire, Magnetic field appears ''' 
    posanph = 0
    neganph = chargespeed
    battery_current = True
    for zf in range(frames3):
    
        plt.clf()
        setplot()
        if zf == 0:
            print ()
            print ('3) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'Closing the switch drives a current from the battery through the wire.')
        Write_Caption(1,'The +ve charges stay fixed in the atoms, but -ve electrons move and form a current.')        
        Write_Caption(2,'The moving charges of the current create a magnetic field that wraps around the wire.') 
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        Draw_Circuit()         
        if zf > 0:
            Draw_Field_Ellipses()
        
        writer.grab_frame()
    
    ''' 4) Charge Moving in Magnetic field, Wire still, -ve charges move to right  ''' 
    posanph = 0
    neganph = chargespeed
    xchg = xcharge
    ychg = ycharge
    for zf in range(frames4):
        plt.clf()
        setplot()
        battery_current = True
        if zf == 0:
            print ()
            print ('4) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'A charge moving to the right through the magnetic field is deflected towards the wire,')
        Write_Caption(1,'at right angles to both the magnetic field and its own direction of motion.')        
        Write_Caption(2,'From this viewpoint, it is a MAGNETIC FIELD that deflects the charge.') 
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        Draw_Circuit()         
        xchg += dxchg
        #ychg -= 1/frames4        
        ychg = ycharge - 2.2*(zf/frames4)**2 
        Draw_Field_Ellipses()
        Draw_Test_Charge()
        
        writer.grab_frame()
    
    ''' ***************************** REVERSE VIEW ***************************** '''
    ''' *********************** Charge Still, Wire Moving  ***************** '''
    ''' 5) Charge still, Wire still, no current  in wire, no field  '''   
    ychg = ycharge
    xchg = (xr+xl)/2
    posanph = 0
    neganph = 0
    battery_current = False
    for zf in range(frames5):
        plt.clf()
        setplot()        
        if zf == 0:
            print ()
            print ('5) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'Now swap viewpoints: the charge stays still and the wire will move to the left.')
        Write_Caption(1,'A stationary charge is unaffected by any magnetic field.')        
        #Write_Caption(2,'')        
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        Draw_Circuit()
        Draw_Test_Charge()    # Now still
        
        writer.grab_frame()
        
    ''' 6) Charge still, Wire speeding up, no current  in wire, no magnetic field  '''   
    ychg = ycharge
    xchg = (xr+xl)/2
    posanph = 0
    neganph = 0
    battery_current = False
    for zf in range(frames6):
        plt.clf()
        setplot()        
        if zf > 0:
            wire_moving = True
        if zf==0:
            print ()
            print ('6) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'The wire starts moving to the left. As it accelerates, special relativity makes it')
        Write_Caption(1,'appear to shrink along its direction of travel. The charges fixed in the wire move')        
        Write_Caption(2,'with it and appear bunched up, but the wire remains electrically neutral overall.')        
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        Draw_Circuit()
        Draw_Test_Charge()    # Now still
        wire_moving = True
        posanph -= chargespeed/frames6        
        neganph = posanph
        half_wire_len = 0.5*(rm-lm)*(1 - (zf+1)*rel_shrink/frames6) # Relativity compression
        xl = pltwid/2 - half_wire_len
        xr = pltwid/2 + half_wire_len
        spposx = chargespacing*(1 - (zf+1)*rel_shrink/frames6)  # Relativity compression
        spnegx = spposx
        
        writer.grab_frame()
        
        
    ''' 7) Charge moving, Wire moving to the left, electric field  '''   
    posanph = - chargespeed
    neganph = 0
    spnegx = chargespacing        # -ve charges back to full spacing
    battery_current = True
    for zf in range(frames7):
        plt.clf()
        setplot()        
        if zf==0:
            print ()
            print ('7) frame',zf, end=',')
        else:
            print (zf, end=',')
        Write_Caption(0,'Now a current drives the -ve charges in the wire to the right at the same speed as the wire')
        Write_Caption(1,'moves to the left. The -ve charges appear motionless and uncompressed, while the +ve charges')        
        Write_Caption(2,'continue moving and appear bunched up. The wire now appears positively charged overall,')        
        Write_Caption(3,'making an ELECTRIC FIELD that attracts the test charge towards the wire.')        
        Draw_Wire()
        Draw_Charges(zf)
        Blanck_Wire_Ends()
        Draw_ZigZag(xl,yc+wdia/2,wdia,ww)
        Draw_ZigZag(xr,yc+wdia/2,wdia,ww)
        #Draw_Field_Ellipses()
        Draw_Circuit()
        Draw_Electric_Field()
        Draw_Test_Charge()    # Now still !
        #ychg -= 1/frames7
        ychg = ycharge - 2.2*(zf/frames7)**2
        writer.grab_frame()

         

plt.savefig('Electric Magnetic.png',dpi=300)


#        # Draw Border Rectangle
#        border = mpatches.Rectangle((0.1, 0.1), pltwid-0.2, pltht-0.2, fill=False, linewidth=1)
#        border = mpatches.Rectangle((lm, bm), pltwid-2*margin, pltht-2*margin, fill=False, linewidth=1)
#        plt.gca().add_patch(border)
